import pandas as pd


def date_split(data):
    data_cp = data.copy()
    data_cp.reset_index(drop=True, inplace=True)
    result = []
    if len(data) == 1:
        result.append([data_cp[0], data_cp[0]])
        return result
    else:
        data_diff = data_cp.diff()
        data_diff.iloc[0] = 1

        # 获取断点列表
        break_point = data_diff[data_diff != 1].index.tolist()
        if len(break_point) == 0:
            result.append([data_cp[0], data_cp[len(data_cp) - 1]])
            return result
        else:
            first_point = 0
            # 根据断点列表切分data
            for i in break_point:
                _data = data_cp[first_point:i]
                _data.reset_index(drop=True, inplace=True)
                result.append([_data[0], _data[len(_data) - 1]])
                first_point = i
            # 最后一个断点后面的数据
            _data = data_cp[first_point:]
            _data.reset_index(drop=True, inplace=True)
            result.append([_data[0], _data[len(_data) - 1]])
    return result


# # main函数
# if __name__ == '__main__':
#     # 新建一个pandas 内容是年份
#     data = pd.DataFrame({'year': [2014, 2015, 2018, 2020]})
#     # 调用函数
#     result = date_split(data['year'])
#     print(result)



import backtrader as bt
import pandas as pd

import zw_backtrader.zw_cerebro as zw
import zw_backtrader.strategy.strategy as zw_strategy

import zw_backtrader.zw_option as z


class data_on(bt.feeds.PandasData):
    lines = ('datetime', 'open', 'high', 'low', 'close', 'pre_close', 'change', 'pct_chg', 'vol', 'amount',)
    params = (
        ('datetime', -1), ('open', -1), ('high', -1), ('low', -1), ('close', -1), ('pre_close', -1), ('change', -1), ('pct_chg', -1), ('vol', -1),
        ('amount', -1),)
    datafields = ['datetime', 'open', 'high', 'low', 'close', 'pre_close', 'change', 'pct_chg', 'vol', 'amount']


#


if "__main__" == __name__:
    # z.lines_params_helper(('trade_date', 'open', 'high', 'low', 'close', 'pre_close', 'change', 'pct_chg', 'vol', 'amount'))

    # 1. 创建日志对象
    logs = zw.log_option()
    # 2. 创建资金对象
    cash = zw.cash_option()
    # 3. 创建cerebro对象
    cerebro = zw.base_cerebro(logs, cash)
    # 4. 导出cerebro核心对象
    cerebro1 = cerebro.get_cerebro()
    # 5. 导入数据
    base_data = pd.read_csv("/Users/summer/PycharmProjects/daily-random-forest-project/data/data.csv")
    test_data = base_data[base_data['ts_code'] == '603019.SH']
    test_data = test_data[test_data['trade_date'] > 20160601]
    aa = test_data[['trade_date', 'open', 'high', 'low', 'close', 'pre_close', 'change', 'pct_chg', 'vol', 'amount']]
    aa.columns = ['datetime', 'open', 'high', 'low', 'close', 'pre_close', 'change', 'pct_chg', 'vol', 'amount']
    aa['datetime'] = aa['datetime'].apply(lambda x: pd.to_datetime(x, format='%Y%m%d'))
    # aa.set_index('datetime', inplace=True)
    data = data_on(dataname=aa)
    # data = bt.feeds.PandasData(dataname=test_data)
    # 6. 添加数据
    cerebro1.adddata(data=data, name='603019.SH')
    # 7. 导入策略
    cerebro1.addstrategy(strategy=zw_strategy.test_strategy)
    # 8. 运行策略
    cerebro1.run()
    # 9. 绘图
    # cerebro1.plot()



import pandas as pd


class trade_detail:
    """
    交易明细
    """

    def __init__(self, buy_date, name, trade_id, sell_date):
        self.name = name
        self.buy_date = buy_date
        self.trade_id = trade_id
        self.sell_date = sell_date



import pandas as pd
import pywencai as wc
import tushare as ts
import os
from kit import num_kit, path_kit
import pathlib

'''
数据函数
'''


# 导入工程
class base_data_loader:
    """
    数据加载器
    通过问财获得股票列表
    通过tushare获得股票数据
    """

    def __init__(self, path="../data/", module="data/", file_name="base_data.csv"):
        self.path = path
        self.module = module
        self.file_name = file_name
        self.module_path = path + module
        self.question = None
        self.start = None
        self.end = None
        self.token = None
        self.symbol_index = dict()  # 股票时间索引
        self.filter = set()  # 股票的过滤集
        self.data = pd.DataFrame()
        path_kit.create_dir(self.module_path)

    def __daily_data(self, symbol, start, end, token):
        """
        获取日线数据
        :param symbol:
        :param start:
        :param end:
        :return:
        """
        api = ts.pro_api(token)
        df = ts.pro_bar(
            ts_code=symbol,
            api=api,
            start_date=str(start) + "0101",
            end_date=str(end) + "1231",
            adjfactor=True,
            asset="E",  # 证券类型 E:股票和交易所基金，I:沪深指数,C:数字货币,FT:期货 FD:基金/O期权/H港股/CB可转债
            freq="D",  # D:日线 W:周 M:月
            adj="hfq",  # 不复权:None 前复权:qfq 后复权:hfq
            retry_count=99  # 重试次数
        )
        return df[::-1]

    def __daily(self, start_date, end_date, token, symbols: list):
        """
        获取日线数据
        :param start_date:
        :param end_date:
        :param symbols:
        :return:
        """
        result = pd.DataFrame()
        if len(symbols) == 0:
            return pd.DataFrame()
        for symbol in symbols:
            df = self.__daily_data(symbol, start_date, end_date, token)
            result = pd.concat([result, df])
        return result

    def filter_symbols(self, symbols: list):
        """
        过滤数据列表
        :param symbols: 以列表的形式填入股票代码
        :return:
        """
        symbols_set = set(symbols)
        self.filter.update(symbols_set)

    def __get_symbols_by_wc(self, question, columns: list):
        """
        股票列表
        通过问财获得股票列表
        """
        result = pd.DataFrame()
        for i in range(self.start, self.end + 1):
            quest = question % (i, i - 1)
            data = wc.get(question=quest, loop=True)
            data = data[columns]
            data = data[~data['股票代码'].isin(self.filter)]
            data['trade_date'] = i
            result = pd.concat([result, data])
        return result

    def get_data(self, question, start, end, token):
        """
        获取总数据集
        优先在本地读取，如果本地没有从互联网获取
        :param token:
        :param end:
        :param start:
        :param question:
        :param data_path: 默认的数据集路径
        :return:
        """

        self.question = question
        self.start = start
        self.end = end

        if os.path.exists(self.module_path + self.file_name):
            print("读取本地数据集")
            self.data = pd.read_csv(self.module_path + self.file_name)
        else:
            print("从互联网获取数据集")
            symbols_list = self.__get_symbols_by_wc(self.question, columns=['股票代码'])
            print("开始遍历")
            for index, symbol in enumerate(symbols_list['股票代码'].unique()):
                print("数据进度百分比:%s" % (index / len(symbols_list['股票代码'].unique()) * 100), end='\r', flush=True)
                self.data = pd.concat([self.data, self.__daily_data(symbol, self.start, self.end, token)])
        # 重置索引
        self.data = self.data.reset_index(drop=True)
        # 将日期转换为字符串
        self.data['trade_date'] = self.data['trade_date'].apply(lambda x: str(x))
        # 将数据里的数字类型转换为float
        self.data = self.data.applymap(lambda x: float(x) if isinstance(x, (int, float)) else x)
        return self.data

    def observe(self, mlflow):
        """
        观察数据集
        :return:
        """
        # 数据报告
        mlflow.log_text("\n".join(self.__data_report()), self.module + "data_report.txt")
        # 获取每个股票的时间个数
        mlflow.log_dict(self.data.groupby('ts_code')['trade_date'].count().to_dict(), self.module + "data_count.txt")
        # 每个数据的缺失日期情况
        mlflow.log_text(self.__row_nan_trade_date(), self.module + "data_trade_nan.txt")

    def save(self, path=None, name=None):
        """
        保存数据集
        :param name:
        :param path:
        :return:
        """
        file_path = self.module_path if path is None else path
        file_name = self.file_name if name is None else name
        self.data.to_csv(file_path + file_name, index=False, encoding='utf-8', )

    def __data_report(self):
        """
        数据报告
        常规基础数据
        :return:
        """
        data = []
        # stringbuffer的数据报告
        data.append("开始日期:%s" % self.start)
        data.append("结束日期:%s" % self.end)
        data.append("数据总量:%s" % len(self.data))
        data.append("数据列数:%s" % len(self.data.columns))
        data.append("数据列名:%s" % self.data.columns.tolist())
        data.append("数据集缺失值:%s" % self.data.isnull().sum())
        return data

    def __row_nan_trade_date(self):
        text = ''
        symbols = self.data["ts_code"].unique()
        trades = self.data["trade_date"].unique()
        for symbol in symbols:
            trade_list = pd.DataFrame(trades)
            trade_list.columns = ['trade_date']
            trade_data = pd.merge(trade_list, self.data[self.data['ts_code'] == symbol], on='trade_date', how='left')
            trade_data = trade_data[trade_data['ts_code'].isnull()]
            if len(trade_data) != 0:
                text = text + symbol + ','
                text = text + ",".join(trade_data['trade_date'].astype('str').tolist()) + '\n'
        return text


# 特征工程
class feature_data_loader:
    """
    数据特征工程
    1.添加特征
    2.观察数据集
    3.保存数据集
    """

    def __init__(self, base_data=None, path="../data/", module="data/"):
        self.data = base_data
        self.features_list = []
        self.result = pd.DataFrame()
        self.path = path
        self.module = module
        self.module_path = path + module

    def add_feature(self, feature):
        """
        添加特征
        :param feature:
        :return:
        """
        for func in feature:
            self.features_list.append(func)
        # 添加特征后，重新初始化数据集

    def obverse(self):
        """
        观察数据集
        :return:
        """
        pass

    def create_execute(self, path=None):
        """
        执行特征工程
        :return:
        """
        file_path = self.module_path if path is None else path
        path_kit.create_dir(file_path + "/feature")
        symbol_list = self.data['ts_code'].unique()
        columns = self.data.columns
        for index, symbol in enumerate(symbol_list):
            print("数据进度百分比:%s" % (index / len(symbol_list) * 100))
            symbol_data = pd.DataFrame(self.data[self.data['ts_code'] == symbol])
            symbol_data.reset_index(drop=True, inplace=True)
            for func in self.features_list:
                func(symbol_data)
            # 将symbol_data 按照旧列和新列分成2个数据集
            symbol_data_left = symbol_data[columns]
            symbol_data_right = symbol_data[symbol_data.columns[~symbol_data.columns.isin(symbol_data_left.columns)]]
            symbol_data_right.applymap(lambda x: round(float(x), 2) if isinstance(x, (int, float)) else x)
            # 将新列数据集和旧列数据集合并
            data = pd.merge(symbol_data_left, symbol_data_right, left_index=True, right_index=True)
            data.dropna(axis=1, how='any', inplace=True)
            data.to_csv(file_path + "/feature/" + symbol + ".csv", index=False, encoding='utf-8')
        return self.result

    def to_execute(self, data, indicator):
        """
        执行特征工程
        :return:
        """
        symbol_list = data['ts_code'].unique()
        columns = data.columns
        for index, symbol in enumerate(symbol_list):
            print("数据进度百分比:%s" % (index / len(symbol_list) * 100))
            symbol_data = pd.DataFrame(data[data['ts_code'] == symbol])
            symbol_data.reset_index(drop=True, inplace=True)
            for func in indicator:
                func(symbol_data)
            # 将symbol_data 按照旧列和新列分成2个数据集
            symbol_data_left = symbol_data[columns]
            symbol_data_right = symbol_data[symbol_data.columns[~symbol_data.columns.isin(symbol_data_left.columns)]]
            symbol_data_right.applymap(lambda x: round(float(x), 2) if isinstance(x, (int, float)) else x)
            # 将新列数据集和旧列数据集合并
            data1 = pd.merge(symbol_data_left, symbol_data_right, left_index=True, right_index=True)
        return data1


# 训练测试工程
class trains_data_loader:

    def __init__(self, path="../data/", module="data/"):
        self.feature_dir = None
        self.path = path
        self.module = module
        self.module_path = path + module
        self.train_X = pd.DataFrame()
        self.train_y = pd.DataFrame()
        self.test_X = pd.DataFrame()
        self.test_y = pd.DataFrame()
        self.drop_column = []

    def load_feature_dir(self, feature_dir):
        self.feature_dir = feature_dir

    def drop_columns(self, columns):
        """
        删除指定列
        :param columns:
        :return:
        """
        for column in columns:
            self.drop_column.append(column)

    def split_by_time(self, trains_start, trains_end, test_start, test_end):
        """
        :param test_end:
        :param test_start:
        :param trains_end:
        :param trains_start:
        :param trans:
        :param start:
        :param end:
        :return:
        """
        self.train_X = pd.DataFrame()
        self.train_y = pd.DataFrame()
        self.test_X = pd.DataFrame()
        self.test_y = pd.DataFrame()

        # trains_start_year = trains_start[0:4]
        # trains_end_year = trains_end[0:4]
        # test_start_year = test_start[0:4]
        # test_end_year = test_end[0:4]
        file_list = os.listdir(self.module_path + self.feature_dir)
        for index, file in enumerate(file_list):
            print(f"读取进度:{(index / len(file_list)) * 100}")
            data = pd.read_csv(self.module_path + self.feature_dir + file, encoding='utf-8')

            trains_x = data[(data['trade_date'] > trains_start) & (data['trade_date'] < trains_end)]
            trains_y = trains_x['flag']
            trains_x = trains_x.drop(self.drop_column, axis=1)
            self.train_X = pd.concat([self.train_X, trains_x])
            self.train_y = pd.concat([self.train_y, trains_y])

            test_X = data[(data['trade_date'] > test_start) & (data['trade_date'] < test_end)]
            test_y = test_X['flag']
            test_X = test_X.drop(self.drop_column, axis=1)
            self.test_X = pd.concat([self.test_X, test_X])
            self.test_y = pd.concat([self.test_y, test_y])

    def obverse(self, mlflow):
        pass
        # mlflow.log_metric("train_label_1", len(self.train_X[self.train_X['flag'] == 1]) / len(self.train_X) * 100)
        # mlflow.log_metric("train_label_0", len(self.train_X[self.train_X['flag'] == 1]) / len(self.train_X) * 100)
        # mlflow.log_metric("train_label_-1", len(self.train_X[self.train_X['flag'] == 1]) / len(self.train_X) * 100)

    def save(self, path=None):
        """
        保存数据集
        :param path:
        :return:
        """

        file_path = self.module_path if path is None else path
        self.train_X.to_pickle(file_path + 'train_X.pkl')
        self.train_y.to_pickle(file_path + 'train_y.pkl')
        self.test_X.to_pickle(file_path + 'test_X.pkl')
        self.test_y.to_pickle(file_path + 'test_y.pkl')


class backtrader_data_loader:

    def __init__(self, path='../data/', module='data/', csv_data='dataset.csv'):
        self.path = path
        self.module = module
        self.data_path = self.path + self.module
        self.csv_data = csv_data
        self.data = pd.DataFrame()

    def get_data(self, start, end):
        """
        复权前的数据+ 复权后的指标
        :param start:
        :param end:
        :return:
        """
        data = pd.read_csv(self.path + self.csv_data)
        self.data = data[(data['trade_date'] > start) & (data['trade_date'] < end)]
        self.data['open'] = round(self.data['open'] / self.data['adj_factor'], 2)
        self.data['high'] = round(self.data['high'] / self.data['adj_factor'], 2)
        self.data['low'] = round(self.data['low'] / self.data['adj_factor'], 2)
        self.data['close'] = round(self.data['close'] / self.data['adj_factor'], 2)
        self.data['amount'] = round(self.data['amount'] / self.data['adj_factor'], 2)
        self.data.drop(['pre_close', 'change', 'pct_chg', 'flag'], axis=1, inplace=True)
        self.data['trade_date'] = self.data['trade_date'].apply(lambda x: pd.to_datetime(x, format='%Y%m%d'))
        self.data = self.data.rename(columns={'trade_date': 'datetime',
                                              'vol': 'volume'
                                              })
        return self.data

    def save(self, module, name):
        path = self.path + module
        if not os.path.exists(path):
            os.mkdir(path)
        self.data.to_csv(path + name, index=False, encoding='utf-8')



import pandas as pd


def get_before_date(date, days=60):
    """
    取得指定日期前60天的日期
    :param date:
    :param days:
    :return:
    """
    date = pd.to_datetime(date)
    date = date - pd.Timedelta(days=days)
    return date.strftime("%Y-%m-%d")

def now(type="%Y%m%d%H%M%S"):
    """
    获取当前时间
    :return:
    """
    return pd.to_datetime("now").strftime(type)


class log_option:
    """
    日志选项
    """

    def __init__(self, log_file_name="log.txt"):
        self.log_file_name = log_file_name


class cash_option:
    """
    股票资金选项
    """

    def __init__(self, cash=50000.0, commission=0.0003, slip=0.005):
        self.cash = cash  # 初始资金
        self.commission = commission  # 交易费率
        self.slip = slip  # 滑点


def lines_params_helper(add_list):
    lines = ()
    params = []
    for col in add_list:
        lines += (col,)
        params.append((col, -1))
    params = tuple(params)
    print("lines = ", lines)
    print("params = ", params)



import backtrader as bt
from zw_backtrader.zw_option import log_option, cash_option


class base_cerebro:
    def __init__(self, logs: log_option, cash: cash_option):
        self.cerebro = bt.Cerebro()
        self.logs = logs
        self.cash_option = cash

    def get_cerebro(self):
        cerebro = self.cerebro
        cerebro.broker.setcash(self.cash_option.cash)  # 设置初始资金
        cerebro.broker.setcommission(self.cash_option.commission)  # 设置交易费率
        cerebro.broker.set_slippage_fixed(self.cash_option.slip)  # 设置固定滑点
        return cerebro  # 返回cerebro核心对象



import backtrader as bt
import joblib
import pandas as pd
import os
from kit import date_kit
import datetime as dt
import zw_backtrader.trade.trade_detail as td


class base_strategy(bt.Strategy):
    """
    策略基类
    同时只能持有一只股票
    卖了以后有10天冷静期
    """

    params = (
        ('model_path', ""),  # 模型路径
        ('log_path', '../data/backtrader/'),  # 日志路径
        ('log_name', f'logs_{date_kit.now()}.log'),  # 日志名称
        ('cold_date', 10)  # 冷静期
    )

    def __init__(self):
        # 买卖列表
        self.position_list = dict()
        # 拦截列表
        self.filter_dict = dict()
        # 日志集合
        self.log_text = []

    def log(self, txt, dt=None):
        """
        Logging function fot this strategy
        :param txt:
        :param dt:
        :return:
        """
        dt = dt or self.datas[0].datetime.date(0)
        self.log_text.append(f'{dt.isoformat()}: {txt}')
        print(f'{dt.isoformat()}: {txt}')

    def start(self):
        """
        策略启动时执行
        :return:
        """
        self.model = joblib.load(self.params.model_path)

    # def prenext(self):
    #     '''策略准备阶段,对应第1根bar ~ 第 min_period-1 根bar'''
    #     # 该函数主要用于等待指标计算，指标计算完成前都会默认调用prenext()空函数
    #     # min_period 就是 __init__ 中计算完成所有指标的第1个值所需的最小时间段
    #     print('prenext函数')
    #
    # def nextstart(self):
    #     '''策略正常运行的第一个时点，对应第 min_period 根bar'''
    #     # 只有在 __init__ 中所有指标都有值可用的情况下，才会开始运行策略
    #     # nextstart()只运行一次，主要用于告知后面可以开始启动 next() 了
    #     # nextstart()的默认实现是简单地调用next(),所以next中的策略逻辑从第 min_period根bar就已经开始执行
    #     print('nextstart函数')

    def next(self):
        # 获取每只股票的信息
        # 并放入模型验证
        for index in range(len(self.datas)):
            self._next_execute(index)

    def _next_execute(self, index):
        symbol = self.datas[index]._name
        date = self.datas[index].datetime.date(0)
        self._sell_excute(date, index)
        if self._predict_excute(index) == 1:
            if self._filter_execute(symbol, date):
                self._buy_excute(symbol, index)
            else:
                pass
                # self.log(f'近期购买,过滤{symbol}')
        else:
            pass
            # self.log(f'预测不买入{symbol}')

    def _filter_execute(self, symbol, date):
        # 过滤持仓
        if self.position_list.get(symbol) is not None:
            # 有持仓 不买入
            return False
        else:
            # 过滤冷静期
            filter_date = self.filter_dict.get(symbol)
            if filter_date is not None:
                if date < filter_date:
                    # 日期小于冷静期 不买入
                    return False
                else:
                    # 日期大于冷静期 买入
                    # 删除冷静期
                    self.filter_dict.pop(symbol)
                    return True
        return True

    def _sell_excute(self, date, index):
        for k, v in self.position_list.items():
            if date > v.sell_date:
                self.sell(data=self.datas[index], size=100, exectype=bt.Order.Market, price=self.datas[index].lines.close[0])

    def _predict_excute(self, index):
        data = self._get_data(datas=self.datas[index], index=index)
        return self.model.predict(data)

    def _buy_excute(self, symbol, index):
        adj_factor = self.datas[index].lines.adj_factor[1]
        # open = round(self.datas[index].lines.open[1] / adj_factor, 2)
        # close = round(self.datas[index].lines.close[0] / adj_factor, 2)
        # high = round(self.datas[index].lines.high[0] / adj_factor, 2)
        # low = round(self.datas[index].lines.low[0] / adj_factor, 2)
        self.buy(data=self.datas[index], size=100)

    def _get_data(self, datas, index):
        # 这个index非常必要熬
        data = pd.DataFrame()
        for v, k in enumerate(datas._colmapping):
            if self.data.data_schema is not None:
                if k not in self.data.data_schema:
                    continue
                exec(f"data.loc[0,'{k}'] = self.datas[index].lines.{k}[0]")
        return data

    def stop(self):
        """
        策略结束时执行
        将日志进行保存
        :return:
        """
        if os.path.exists(self.params.log_path) is False:
            os.mkdir(self.params.log_path)

        with open(self.params.log_path + self.params.log_name, mode='w') as f:
            f.write('\n'.join(self.log_text))

    def notify_order(self, order):
        """
        订单状态变化时执行
        :param order:
        :return:
        """

        # 判断订单的状态
        if order.status in [order.Submitted, order.Accepted]:
            return
        # 判断订单是否完成
        if order.status in [order.Completed]:
            if order.isbuy():
                self.log(f'买入{order.data._name},价格:{order.executed.price},成本:{order.executed.value},手续费:{order.executed.comm}')
                self.position_list.update({order.data._name: td.trade_detail(order.data.datetime.date(0), order.data._name, trade_id=order.ref,
                                                                             sell_date=order.data.datetime.date(3))})

            else:
                self.log(f'卖出{order.data._name},价格:{order.executed.price},成本:{order.executed.value},手续费:{order.executed.comm}')
                self.position_list.pop(order.data._name)
                self.filter_dict[order.data._name] = order.data.datetime.date(0) + dt.timedelta(days=self.params.cold_date)
        elif order.status in [order.Canceled, order.Margin, order.Rejected]:
            self.log(f'订单失败: {order.data._name} 状态: {order.getstatusname()}')

    def notify_trade(self, trade):
        """
        交易状态变化时执行
        :param trade:
        :return:
        """
        if not trade.isclosed:
            return
        self.log(f'利润,毛利:{trade.pnl},净利:{trade.pnlcomm}')

    # def notify_cashvalue(self, cash, value):
    #     """
    #     资金变化时执行
    #     :param cash:
    #     :param value:
    #     :return:
    #     """
    #     self.log(f'资金:{cash},市值:{value}')

    # def notify_fund(self, cash, value, fundvalue, shares):
    #     """
    #     资金变化时执行
    #     :param cash:
    #     :param value:
    #     :param fundvalue:
    #     :param shares:
    #     :return:
    #     """
    #     self.log(f'资金:{cash},市值:{value},净值:{fundvalue},持仓:{shares}')

# class test_strategy(bt.Strategy):
#     """
#     策略基类
#     """
#
#     def __init__(self):
#         # 买卖列表
#         self.buying_list = []
#         self.selling_list = []
#
#     def next(self):
#         for index in range(len(self.datas)):
#             data = self._get_data(datas=self.datas[index], index=index)
#             print(data)
#
#             # 进行预测
#
#     def _get_data(self, datas, index):
#         data = pd.DataFrame()
#         for v, k in enumerate(datas._colmapping):
#             if k is None:
#                 continue
#             exec(f"data.loc[0,'{k}'] = self.datas[index].lines.{k}[0]")
#         return data
